from pydantic import Field

MAP_FIELD_FLAG = 'map_field'


def MapField(*args, **kwargs):
    """
    Has the same functionality as pydantic `Field` but adds a flag to mark the field as map attribute that
    responsible for the decision which map config to load.
    """
    if kwargs.get(MAP_FIELD_FLAG):
        raise ValueError(f'Cannot use "{MAP_FIELD_FLAG}" key inside `MapField()`.')
    return Field(*args, **kwargs, **{MAP_FIELD_FLAG: True})





from typing import Optional, Callable, List

from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader


class LoaderManager:
    def __init__(
            self,
            settings_obj: object,
            source_priority: List[str],
            init_settings_callable: Optional[Callable] = None,
            env_settings_callable: Optional[Callable] = None,
            file_secret_settings_callable: Optional[Callable] = None,
            loaders: Optional[List[SourceLoader]] = None,
    ):
        self.settings_obj = settings_obj
        self.init_settings_callable = init_settings_callable
        self.env_settings_callable = env_settings_callable
        self.file_secret_settings_callable = file_secret_settings_callable
        self.loaders = loaders or []
        self.source_priority = source_priority
        self.all_loaded_fields = {}
        self.full_fields = {}

    def load_all(self):
        loaders_dict = {loader.NAME: loader for loader in self.loaders}
        for source in self.source_priority:
            if source is ConfigSource.map:
                continue
            self.all_loaded_fields[source] = {
                field.name: field for field in loaders_dict[source].load_fields(settings=self.settings_obj)
            }

        # Load the map config last.
        if ConfigSource.map in self.source_priority:
            self.all_loaded_fields[ConfigSource.map] = {
                field.name: field for field in loaders_dict[ConfigSource.map].load_fields(settings=self.settings_obj)
            }

        # As needed in `customise_sources`, Create a tuple of callbacks that return dict of fields for every source.
        # The first callable has the highest priority and so on.
        source_fields_callables = (lambda s: self.all_loaded_fields[source] for source in self.source_priority)

        return source_fields_callables



from typing import List, Callable

from pydantic import BaseSettings

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader


class InitSourceLoader(SourceLoader):
    NAME = ConfigSource.init

    def __init__(self, init_settings_callable: Callable, **kwargs):
        super().__init__(**kwargs)
        self.init_settings_callable = init_settings_callable

    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        fields = self.init_settings_callable(settings=settings)
        full_fields = [
            ConfigField(
                name=key, value=value, source_name=ConfigSource.init, source_type=ConfigSource.init,
                source_location=self.specs.creation_path
            ) for key, value in fields.items()
        ]
        return full_fields



from typing import Callable, List

from pydantic import BaseSettings

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader
from confident.utils import convert_field_value


class EnvSourceLoader(SourceLoader):
    NAME = ConfigSource.env_var

    def __init__(self, env_settings_callable: Callable, **kwargs):
        super().__init__(**kwargs)
        self.env_settings_callable = env_settings_callable

    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        """
        Finds and loads requested settings fields from environment variables into a dictionary.
        """
        fields = self.env_settings_callable(settings=settings)
        full_fields = [
            ConfigField(
                name=key, value=convert_field_value(settings=settings, field_name=key, origin_value=value),
                origin_value=value,
                source_name=key,
                source_type=ConfigSource.env_var,
                source_location=key
            ) for key, value in fields.items()
        ]
        return full_fields
